#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/module.h>

#define PREFIX KBUILD_MODNAME ": "

struct super_miscdevice {
	struct miscdevice misc;
	int registered;
};

static int super_misc_register(struct super_miscdevice *m)
{
	int ret;
	if ((ret = misc_register(&m->misc)) != 0)
		return ret;
	m->registered = 1;
	return 0;
}

static void super_misc_deregister(struct super_miscdevice *m)
{
	if (m->registered) {
		misc_deregister(&m->misc);
		m->registered = 0;
	}
	return;
}

static int bug_open(struct inode *inode, struct file *filp)
{
	BUG();
	return -EIO;
}

static const struct file_operations bug_fops = {
	.open  = bug_open,
	.owner = THIS_MODULE,
};

static struct super_miscdevice bug_misc = {
	.misc  = {
		.minor = MISC_DYNAMIC_MINOR,
		.name  = "xxbug",
		.fops  = &bug_fops,
	},
};

static int oops_open(struct inode *inode, struct file *filp)
{
	++*(int *)NULL;
	return -EIO;
}

static const struct file_operations oops_fops = {
	.open  = oops_open,
	.owner = THIS_MODULE,
};

static struct super_miscdevice oops_misc = {
	.misc = {
		.minor = MISC_DYNAMIC_MINOR,
		.name  = "xxoops",
		.fops  = &oops_fops,
	},
};

static int panic_open(struct inode *inode, struct file *filp)
{
	panic(PREFIX "PANIC triggered\n");
	return -EIO;
}

static const struct file_operations panic_fops = {
	.open  = panic_open,
	.owner = THIS_MODULE,
};

static struct super_miscdevice panic_misc = {
	.misc = {
		.minor = MISC_DYNAMIC_MINOR,
		.name  = "xxpanic",
		.fops  = &panic_fops,
	},
};

static int runaway_open(struct inode *inode, struct file *filp)
{
	unsigned long end = jiffies + 10 * HZ;
	printk(KERN_WARNING PREFIX "Hang");
	while(jiffies < end)
		;
	printk("over\n");
	return -EIO;
}

static const struct file_operations runaway_fops = {
	.open  = runaway_open,
	.owner = THIS_MODULE,
};

static struct super_miscdevice runaway_misc = {
	.misc = {
		.minor = MISC_DYNAMIC_MINOR,
		.name  = "xxrunaway",
		.fops  = &runaway_fops,
	},
};

static __init int explode_init(void)
{
	int ret;
	if ((ret = super_misc_register(&bug_misc)) != 0 ||
	    (ret |= super_misc_register(&oops_misc)) ||
	    (ret |= super_misc_register(&panic_misc)) ||
	    (ret |= super_misc_register(&runaway_misc))) {
		cleanup_module();
		return ret;
	}
	return 0;
}

static __exit void explode_exit(void)
{
	super_misc_deregister(&bug_misc);
	super_misc_deregister(&oops_misc);
	super_misc_deregister(&panic_misc);
	super_misc_deregister(&runaway_misc);
	return;
}

module_init(explode_init);
module_exit(explode_exit);

MODULE_DESCRIPTION("Provide problems");
MODULE_AUTHOR("Jan Engelhardt <jengelh [at] gmx de>");
MODULE_LICENSE("Public Domain");
